"""mdpo I/O utilities."""

import glob
import os
from contextlib import contextmanager


@contextmanager
def environ(**env):
    """Overwrite temporarily some environment variables."""
    original_env = {key: os.getenv(key) for key in env}
    os.environ.update(env)
    try:
        yield
    finally:
        for key, value in original_env.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value


def filter_paths(filepaths, ignore_paths=frozenset()):
    """Filter a list of paths removing those defined in other list of paths.

    The paths to filter can be defined in the list of paths to ignore in
    several forms:

    - The same string.
    - Only the file name.
    - Only their direct directory name.
    - Their direct directory full path.

    Args:
        filepaths (list): Set of source paths to filter.
        ignore_paths (list): Paths that must not be included in the response.

    Returns:
        list: Non filtered paths ordered alphabetically.
    """
    response = []
    for filepath in filepaths:
        # ignore by filename
        if os.path.basename(filepath) in ignore_paths:
            continue
        # ignore by dirname
        if os.path.basename(os.path.dirname(filepath)) in ignore_paths:
            continue
        # ignore by filepath
        if filepath in ignore_paths:
            continue
        # ignore by dirpath (relative or absolute)
        if (os.sep).join(filepath.split(os.sep)[:-1]) in ignore_paths:
            continue
        response.append(filepath)
    response.sort()
    return response


def to_file_content_if_is_file(value, encoding='utf-8'):
    """Check if the value passed is a file path or string content.

    If is a file, reads its content and returns it, otherwise returns
    the string passed as is.

    Args:
        value (str): Value to check if is a filepath or content.
        encoding (str): Expected file encoding, if is a file.

    Returns:
        str: File content if ``value`` is an existing file or ``value`` as is.
    """
    if os.path.isfile(value):
        with open(value, encoding=encoding) as f:
            value = f.read()
    return value


def to_files_or_content(value):
    """File path/glob/content disambiguator.

    Check if the value passed is a glob, a set of files in a list or is string
    content.

    Args:
        value (str): Value to check.

    Returns:
        tuple: Two values being the first a boolean that indicates if ``value``
        is a list of files (``True``) or string content (``False``) and the
        second value is the content, which could be an iterator (if a glob or
        a list of files is passed or a string).
    """
    try:
        parsed = glob.glob(value)
    except TypeError:
        # inferes list
        return (True, value)
    except Exception as err:
        # some strings like '[s-m]' will produce
        if err.__module__ == 're' and err.__class__.__name__ == 'error':
            return (False, value)
        raise err
    if not parsed:
        # assumes it is content
        return (False, value)
    return (True, parsed)


def save_file_checking_file_changed(filepath, content, encoding='utf-8'):
    """Save a file checking if the content has changed.

    Args:
        filepath (:py:class:`polib.POFile`): Path to the file to save.
        content (str): Content to save.
        encoding (str): Encoding to use when saving the file.

    Returns:
        bool: If the PO file content has been changed.
    """
    if not os.path.isfile(filepath):
        with open(filepath, 'w', encoding=encoding) as f:
            f.write(content)
        return True

    with open(filepath, encoding=encoding) as f:
        prev_content = f.read()
    changed = prev_content != content
    if changed:
        with open(filepath, 'w', encoding=encoding) as f:
            f.write(content)
    return changed



"""Compatibilities for different versions of Python."""

import sys


if sys.version_info < (3, 10):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata

__all__ = (
    'importlib_metadata',
)



"""Markdown to PO files extractor according to mdpo specification."""

import contextlib
import glob
import os

import md4c
import polib

from mdpo.command import (
    normalize_mdpo_command_aliases,
    parse_mdpo_html_command,
)
from mdpo.event import add_debug_events, parse_events_kwarg, raise_skip_event
from mdpo.io import (
    filter_paths,
    save_file_checking_file_changed,
    to_files_or_content,
)
from mdpo.md import parse_link_references
from mdpo.md4c import (
    DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS,
    READABLE_BLOCK_NAMES,
)
from mdpo.po import (
    find_entry_in_entries,
    mark_not_found_entries_as_obsoletes,
    po_escaped_string,
    remove_not_found_entries,
)
from mdpo.text import min_not_max_chars_in_a_row, parse_wrapwidth_argument


class Md2Po:
    """Markdown to PO files extractor.

    This class is where all the extraction process is carried out.
    If you are executing custom extraction events, you may want to
    read the documentation about the properties of this class to
    properly control the internal state of the parser.

    Example:
        If you want to extract all "Foo" messages as "Bar",
        regardless of the content of the Markdown input,
        you could do something like:

        .. code-block:: python

           def transform_foo(self, block, text):
               if text == 'Foo':
                   self.current_msgid = 'Bar'  # self is Md2Po
                   return False

           markdown_to_pofile('Foo', events={'text': transform_foo})

    The public internal properties of this class are documented
    below:
    """

    __slots__ = {
        'filepaths',
        'content',
        'pofile',
        'po_filepath',
        'msgstr',
        'found_entries',
        'disabled_entries',
        'ignore_msgids',
        'command_aliases',
        'mark_not_found_as_obsolete',
        'preserve_not_found',
        'extensions',
        'plaintext',
        'include_codeblocks',
        'metadata',
        'events',
        'obsoletes',

        'location',
        '_current_top_level_block_number',
        '_current_top_level_block_type',
        '_current_markdown_filepath',

        # Public class properties
        'current_msgid',
        'current_tcomment',
        'current_msgctxt',
        'link_references',
        'disable',
        'disable_next_block',
        'enable_next_block',
        'include_next_codeblock',
        'disable_next_codeblock',

        '_saved_files_changed',

        '_enterspan_replacer',
        '_leavespan_replacer',

        'bold_start_string',
        'bold_end_string',
        'italic_start_string',
        'italic_start_string_escaped',
        'italic_end_string',
        'italic_end_string_escaped',
        'code_start_string',
        'code_start_string_escaped',
        'code_end_string',
        'code_end_string_escaped',
        'strikethrough_start_string',
        'strikethrough_end_string',
        'latexmath_start_string',
        'latexmath_end_string',
        'latexmathdisplay_start_string',
        'latexmathdisplay_end_string',
        'wikilink_start_string',
        'wikilink_end_string',
        'underline_start_string',
        'underline_end_string',

        '_inside_uspan',
        '_inside_htmlblock',
        '_inside_codeblock',
        '_inside_pblock',
        '_inside_aspan',
        '_inside_liblock',
        '_inside_codespan',
        '_inside_olblock',
        '_inside_hblock',
        '_quoteblocks_deep',
        '_codespan_start_index',
        '_codespan_backticks',
        '_current_aspan_text',
        '_current_aspan_ref_target',
        '_current_wikilink_target',
        '_inside_latexmath_display',
        '_inside_latexmath_display_text',
        '_current_imgspan',
        '_uls_deep',
    }

    def __init__(self, files_or_content, **kwargs):
        is_glob, files_or_content = to_files_or_content(files_or_content)
        if is_glob:
            filepaths = []
            for globpath in files_or_content:
                filepaths.extend(glob.glob(globpath))
            self.filepaths = filter_paths(
                filepaths,
                ignore_paths=kwargs.get('ignore', []),
            )
        else:
            self.content = files_or_content

        #: :py:class:`polib.POFile` PO file object representing
        #: the extracted content.
        self.pofile = None

        #: str: PO file path to which the content will be extracted.
        self.po_filepath = None

        #: str: Default msgstr used if the current is not found
        #: inside the previous content of the specified PO file.
        self.msgstr = kwargs.get('msgstr', '')

        #: list: Extracted entries.
        self.found_entries = []

        #: list: Not extracted entries because the extractor
        #: has been disabled while processing them.
        self.disabled_entries = []

        #: list(str): MD4C extensions used to parse the content.
        #: See all available in :doc:`/dev/reference/mdpo.md4c`.
        self.extensions = kwargs.get(
            'extensions',
            DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS,
        )

        #: dict: Custom events excuted during the parsing while
        #: extracting content.
        self.events = parse_events_kwarg(kwargs.get('events') or {})
        if kwargs.get('debug'):
            add_debug_events('md2po', self.events)

        #: bool: If there are entries in the PO file that
        #: are obsolete and not found in the current extraction.
        self.obsoletes = False

        #: str: The msgid being currently built for the next
        #: message entry. Keep in mind that, if you are executing
        #: an event that will be followed by an span one
        #: (``enter_span`` or ``exit_span``), the content of the
        #: msgid will change before save it.
        self.current_msgid = ''

        #: str: Translator comment that will be saved in the next
        #: message.
        self.current_tcomment = None

        #: str: Context message that will be saved in the next
        #: message.
        self.current_msgctxt = None

        #: bool: Indicates if the extractor is currently disabled,
        #: which happens after a ``<!-- mdpo-disable -->`` command
        #: is found, before any subsecuents ``<!-- mdpo-enable -->``
        #: commands.
        self.disable = False

        #: bool: Indicates if the next block will be extracted.
        self.disable_next_block = False

        #: bool: Indicates if the next block will be extracted
        #: when the extractor is disabled (``disable is True``).
        self.enable_next_block = False

        #: bool: Indicates if the next codeblock will be extracted
        #: when `include_codeblocks` is enabled.
        self.include_next_codeblock = False

        #: bool: Indicates if the next codeblock will be extracted
        #: when `include_codeblocks` is disabled.
        self.disable_next_codeblock = False

        #: bool: Extract code blocks
        self.include_codeblocks = kwargs.get('include_codeblocks', False)

        #: The msgids to ignore for extraction
        self.ignore_msgids = kwargs.get('ignore_msgids', [])

        self.command_aliases = normalize_mdpo_command_aliases(
            kwargs.get('command_aliases') or {},
        )

        self.mark_not_found_as_obsolete = kwargs.get(
            'mark_not_found_as_obsolete', True,
        )
        self.preserve_not_found = kwargs.get('preserve_not_found', True)

        self.plaintext = kwargs.get('plaintext', False)

        self._saved_files_changed = (
            False if kwargs.get('_check_saved_files_changed') else None
        )

        self.metadata = kwargs.get('metadata') or {}

        self.location = kwargs.get('location', True)
        # "top level" here because blocks inside blocks are not taken into
        # account for locations
        self._current_top_level_block_number = 0
        self._current_top_level_block_type = None
        self._current_markdown_filepath = None
        self._inside_latexmath_display = False

        if not self.plaintext:
            self.bold_start_string = kwargs.get('bold_start_string', '**')
            self.bold_end_string = kwargs.get('bold_end_string', '**')

            self.italic_start_string = kwargs.get('italic_start_string', '*')
            self.italic_start_string_escaped = po_escaped_string(
                self.italic_start_string,
            )

            self.italic_end_string = kwargs.get('italic_end_string', '*')
            self.italic_end_string_escaped = po_escaped_string(
                self.italic_end_string,
            )

            # codespans are built by a indetermined number of 'x' characters
            # so we take only the first
            self.code_start_string = kwargs.get('code_start_string', '`')[0]
            self.code_start_string_escaped = po_escaped_string(
                self.code_start_string,
            )

            self.code_end_string = kwargs.get('code_end_string', '`')[0]
            self.code_end_string_escaped = po_escaped_string(
                self.code_end_string,
            )

            _include_xheader = kwargs.get('xheader', False)

            if _include_xheader:
                from mdpo.compat import importlib_metadata
                version = importlib_metadata.version('mdpo')
                self.metadata.update({'X-Generator': f'mdpo v{version}'})

            self._enterspan_replacer = {
                md4c.SpanType.STRONG.value: self.bold_start_string,
                md4c.SpanType.EM.value: self.italic_start_string,
                md4c.SpanType.CODE.value: self.code_start_string,
            }

            self._leavespan_replacer = {
                md4c.SpanType.STRONG.value: self.bold_end_string,
                md4c.SpanType.EM.value: self.italic_end_string,
                md4c.SpanType.CODE.value: self.code_end_string,
            }

            if 'strikethrough' in self.extensions:
                self.strikethrough_start_string = kwargs.get(
                    'strikethrough_start_string', '~~',
                )
                self._enterspan_replacer[md4c.SpanType.DEL.value] = \
                    self.strikethrough_start_string

                self.strikethrough_end_string = kwargs.get(
                    'strikethrough_end_string', '~~',
                )
                self._leavespan_replacer[md4c.SpanType.DEL.value] = \
                    self.strikethrough_end_string

            if 'latex_math_spans' in self.extensions:
                self.latexmath_start_string = kwargs.get(
                    'latexmath_start_string', '$',
                )
                self._enterspan_replacer[md4c.SpanType.LATEXMATH.value] = \
                    self.latexmath_start_string

                self.latexmath_end_string = kwargs.get(
                    'latexmath_end_string', '$',
                )
                self._leavespan_replacer[md4c.SpanType.LATEXMATH.value] = \
                    self.latexmath_end_string

                self.latexmathdisplay_start_string = kwargs.get(
                    'latexmathdisplay_start_string', '$$',
                )
                self._enterspan_replacer[
                    md4c.SpanType.LATEXMATH_DISPLAY.value
                ] = self.latexmathdisplay_start_string

                self.latexmathdisplay_end_string = kwargs.get(
                    'latexmathdisplay_end_string', '$$',
                )
                self._leavespan_replacer[
                    md4c.SpanType.LATEXMATH_DISPLAY.value
                ] = self.latexmathdisplay_end_string

                self._inside_latexmath_display_text = ''

            if 'wikilinks' in self.extensions:
                self.wikilink_start_string = kwargs.get(
                    'wikilink_start_string', '[[',
                )
                self.wikilink_end_string = kwargs.get(
                    'wikilink_end_string', ']]',
                )

                self._enterspan_replacer[md4c.SpanType.WIKILINK.value] = \
                    self.wikilink_start_string
                self._leavespan_replacer[md4c.SpanType.WIKILINK.value] = \
                    self.wikilink_end_string

            if 'underline' in self.extensions:
                # underline text is standarized with double '_'
                self.underline_start_string = kwargs.get(
                    'underline_start_string', '__',
                )
                self._enterspan_replacer[md4c.SpanType.U.value] = \
                    self.underline_start_string

                self.underline_end_string = kwargs.get(
                    'underline_end_string', '__',
                )
                self._leavespan_replacer[md4c.SpanType.U.value] = \
                    self.underline_end_string

            # optimization to skip checking for
            # ``self.md4c_generic_parser_kwargs.get('underline')``
            # inside spans
            self._inside_uspan = False

        self._inside_htmlblock = False
        self._inside_codeblock = False
        self._inside_pblock = False
        self._inside_liblock = False
        self._inside_hblock = False
        self._inside_olblock = False
        self._inside_codespan = False

        self._quoteblocks_deep = 0
        self._uls_deep = 0

        self._codespan_start_index = None
        self._codespan_backticks = None

        self._inside_aspan = False
        self._current_aspan_text = ''
        # indicates the target of the current link, which is referenced and
        # extracted without using MD4C, so we can preserve it as referenced
        self._current_aspan_ref_target = None

        self.link_references = None
        self._current_wikilink_target = None
        self._current_imgspan = {}

    def _save_msgid(
        self,
        msgid,
        msgstr='',
        tcomment=None,
        msgctxt=None,
        fuzzy=False,
    ):
        if msgid in self.ignore_msgids:
            return
        entry = polib.POEntry(
            msgid=msgid,
            msgstr=msgstr,
            comment=tcomment,
            msgctxt=msgctxt,
            flags=[] if not fuzzy else ['fuzzy'],
        )

        occurrence = None
        if self.location and self._current_markdown_filepath:
            # here could happen a KeyError if someone has aborted an ,
            # enter event, in which case we do not have access to the
            # block type because ``self._current_top_level_block_type is None``
            current_block_name = READABLE_BLOCK_NAMES[
                self._current_top_level_block_type
            ]
            # TODO: when all tests added and the location feature is fully
            #       tested, we could ignore the KeyError event keeping an
            #       incomplete occurrence place string, and raising a warning
            #       if the user has configured an `enter_block` or
            #       `leave_block` event remembering that
            #       `_current_top_level_block_number` and
            #       `_current_top_level_block_type` properties must be handled
            #       accordingly

            occurrence = (
                self._current_markdown_filepath,
                (
                    f'block {self._current_top_level_block_number}'
                    f' ({current_block_name})'
                ),
            )

            if occurrence not in entry.occurrences:
                entry.occurrences.append(occurrence)

        _equal_entry = find_entry_in_entries(
            entry,
            self.pofile,
            compare_obsolete=False,
            compare_msgstr=False,
            compare_occurrences=False,
        )

        if _equal_entry and _equal_entry.msgstr:
            entry.msgstr = _equal_entry.msgstr
            if _equal_entry.fuzzy and not entry.fuzzy:
                entry.flags.append('fuzzy')
        if entry not in self.pofile:
            self.pofile.append(entry)
        self.found_entries.append(entry)

    def _save_current_msgid(
            self,
            msgstr='',
            fuzzy=False,
    ):
        # raise 'msgid' event
        if raise_skip_event(
            self.events,
            'msgid',
            self,
            self.current_msgid,
            msgstr,
            self.current_msgctxt,
            self.current_tcomment,
            ['fuzzy'] if fuzzy else [],
        ):
            return

        if self.current_msgid:
            if (not self.disable_next_block and not self.disable) or \
                    self.enable_next_block:
                self._save_msgid(
                    self.current_msgid,
                    msgstr=msgstr or self.msgstr,
                    msgctxt=self.current_msgctxt,
                    tcomment=self.current_tcomment,
                    fuzzy=fuzzy,
                )
            else:
                self.disabled_entries.append(
                    polib.POEntry(
                        msgid=self.current_msgid,
                        msgstr=msgstr or self.msgstr,
                        msgctxt=self.current_msgctxt,
                        tcomment=self.current_tcomment,
                        flags=['fuzzy'] if fuzzy else [],
                    ),
                )
        self.disable_next_block = False
        self.enable_next_block = False
        self.current_msgid = ''
        self.current_tcomment = None
        self.current_msgctxt = None

    def command(self, mdpo_command, comment, original_command):
        # raise 'command' event
        if raise_skip_event(
            self.events,
            'command',
            self,
            mdpo_command,
            comment,
            original_command,
        ):
            return

        if mdpo_command in (
            'mdpo-disable-next-block',
            'mdpo-disable-next-line',
        ):
            self.disable_next_block = True
        elif mdpo_command == 'mdpo-disable':
            self.disable = True
        elif mdpo_command == 'mdpo-enable':
            self.disable = False
        elif mdpo_command in (
            'mdpo-enable-next-block',
            'mdpo-enable-next-line',
        ):
            self.enable_next_block = True
        elif mdpo_command == 'mdpo-include-codeblock':
            self.include_next_codeblock = True
        elif mdpo_command == 'mdpo-disable-codeblock':
            self.disable_next_codeblock = True
        elif mdpo_command == 'mdpo-disable-codeblocks':
            self.include_codeblocks = False
        elif mdpo_command == 'mdpo-include-codeblocks':
            self.include_codeblocks = True
        elif mdpo_command == 'mdpo-translator':
            if not comment:
                raise ValueError(
                    'You need to specify a string for the'
                    ' extracted comment with the command'
                    f" '{original_command}'.",
                )
            self.current_tcomment = comment
        elif mdpo_command == 'mdpo-context':
            if not comment:
                raise ValueError(
                    'You need to specify a string for the'
                    f" context with the command '{original_command}'.",
                )
            self.current_msgctxt = comment
        elif mdpo_command == 'mdpo-include':
            if not comment:
                raise ValueError(
                    'You need to specify a message for the'
                    ' comment to include with the command'
                    f" '{original_command}'.",
                )
            self.current_msgid = comment
            self._save_current_msgid()

    def _process_command(self, text):
        original_command, comment = parse_mdpo_html_command(text)
        if original_command is None:
            return

        try:
            command = self.command_aliases[original_command]
        except KeyError:  # not custom command
            command = original_command

        # process solved command
        self.command(command, comment, original_command)

    def enter_block(self, block, details):
        # raise 'enter_block' event
        if raise_skip_event(self.events, 'enter_block', self, block, details):
            return

        if block is md4c.BlockType.P:
            self._inside_pblock = True
            if not any([
                self._inside_hblock,
                self._uls_deep,
                self._quoteblocks_deep,
                self._inside_olblock,
            ]):
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.P.value
        elif block is md4c.BlockType.CODE:
            self._inside_codeblock = True
            if not any([
                self._quoteblocks_deep,
                self._uls_deep,
                self._inside_olblock,
            ]):
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.CODE.value
        elif block is md4c.BlockType.LI:
            self._inside_liblock = True
        elif block is md4c.BlockType.UL:
            self._uls_deep += 1
            if self._uls_deep > 1 or self._inside_olblock:
                # changing UL deeep
                self._save_current_msgid()
            elif not any([
                self._quoteblocks_deep,
                self._inside_olblock,
            ]):
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.UL.value
        elif block is md4c.BlockType.H:
            self._inside_hblock = True
            if not any([
                self._quoteblocks_deep,
                self._uls_deep,
                self._inside_olblock,
            ]):
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.H.value
        elif block is md4c.BlockType.QUOTE:
            self._quoteblocks_deep += 1
            if self._inside_liblock:
                self._save_current_msgid()
            if self._quoteblocks_deep == 1:
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.QUOTE.value
        elif block is md4c.BlockType.OL:
            if not any([
                self._quoteblocks_deep,
                self._uls_deep,
                self._inside_olblock,
            ]):
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.OL.value

            if self._inside_olblock or self._uls_deep:
                self._save_current_msgid()
            self._inside_olblock = True
        elif block is md4c.BlockType.HTML:
            self._inside_htmlblock = True
            if not any([
                self._quoteblocks_deep,
                self._inside_olblock,
                self._uls_deep,
            ]):
                self._current_top_level_block_number += 1
                self._current_top_level_block_type = md4c.BlockType.HTML.value
        elif block is md4c.BlockType.TABLE and not any([
            self._quoteblocks_deep,
            self._inside_olblock,
            self._uls_deep,
        ]):
            self._current_top_level_block_number += 1
            self._current_top_level_block_type = md4c.BlockType.TABLE.value

    def leave_block(self, block, details):
        # raise 'leave_block' event
        if raise_skip_event(self.events, 'leave_block', self, block, details):
            return

        if block is md4c.BlockType.CODE:
            self._inside_codeblock = False
            if not self.disable_next_codeblock and (
                self.include_codeblocks or self.include_next_codeblock
            ):
                self._save_current_msgid()
            self.include_next_codeblock = False
            self.disable_next_codeblock = False
        elif block is md4c.BlockType.HTML:
            self._inside_htmlblock = False
        else:
            if block is md4c.BlockType.P:
                self._inside_pblock = False
            elif block is md4c.BlockType.LI:
                self._inside_liblock = True
            elif block is md4c.BlockType.UL:
                self._uls_deep -= 1
            elif block is md4c.BlockType.H:
                self._inside_hblock = False
            elif block is md4c.BlockType.QUOTE:
                self._quoteblocks_deep -= 1
            elif block is md4c.BlockType.OL:
                self._inside_olblock = False
            self._save_current_msgid()

    def enter_span(self, span, details):
        # raise 'enter_span' event
        if raise_skip_event(self.events, 'enter_span', self, span, details):
            return

        if (span is md4c.SpanType.IMG or span is md4c.SpanType.A) and \
                details['title']:
            self._save_msgid(details['title'][0][1])

    def not_plaintext_enter_span(self, span, details):
        # raise 'enter_span' event
        if raise_skip_event(self.events, 'enter_span', self, span, details):
            return

        # underline spans for double '_' character enters two times
        if not self._inside_uspan:
            if self._inside_aspan:  # span inside link text
                with contextlib.suppress(KeyError):
                    self._current_aspan_text += self._enterspan_replacer[
                        span.value
                    ]
            elif span is not md4c.SpanType.LATEXMATH_DISPLAY:
                with contextlib.suppress(KeyError):
                    self.current_msgid += (
                        self._enterspan_replacer[span.value]
                    )

        if span is md4c.SpanType.A:
            # here resides the logic of discover if the current link
            # is referenced
            if self.link_references is None:
                self.link_references = parse_link_references(self.content)

            self._inside_aspan = True

            current_aspan_href = details['href'][0][1]
            self._current_aspan_ref_target = None

            if details['title']:
                current_aspan_title = details['title'][0][1]
                for target, href, title in self.link_references:
                    if (
                        href == current_aspan_href
                        and title == current_aspan_title
                    ):
                        self._current_aspan_ref_target = target
                        break
            else:
                for target, href, _ in self.link_references:
                    if href == current_aspan_href:
                        self._current_aspan_ref_target = target
                        break

        elif span is md4c.SpanType.CODE:
            self._inside_codespan = True

            # entering a code span, literal backticks encountered inside
            # will be escaped
            #
            # save the index char of the opening backtick
            self._codespan_start_index = len(self.current_msgid) - 1
        elif span is md4c.SpanType.IMG:
            if self.link_references is None:
                self.link_references = parse_link_references(self.content)

            self._current_imgspan['src'] = details['src'][0][1]
            self._current_imgspan['title'] = '' if not details['title'] \
                else details['title'][0][1]
            self._current_imgspan['text'] = ''
        elif span is md4c.SpanType.U:
            self._inside_uspan = True
        elif span is md4c.SpanType.WIKILINK:
            self._current_wikilink_target = details['target'][0][1]
        elif span is md4c.SpanType.LATEXMATH_DISPLAY:
            self._inside_latexmath_display = True

    def leave_span(self, span, details):
        # raise 'leave_span' event
        if raise_skip_event(self.events, 'leave_span', self, span, details):
            return

    def not_plaintext_leave_span(self, span, details):
        # raise 'leave_span' event
        if raise_skip_event(self.events, 'leave_span', self, span, details):
            return

        if not self._inside_uspan:
            if span is md4c.SpanType.WIKILINK:
                self.current_msgid += self._current_wikilink_target
                self._current_wikilink_target = None
            if self._inside_aspan:  # span inside link text
                with contextlib.suppress(KeyError):
                    self._current_aspan_text += self._leavespan_replacer[
                        span.value
                    ]
            elif not self._inside_latexmath_display:
                with contextlib.suppress(KeyError):
                    self.current_msgid += (
                        self._leavespan_replacer[span.value]
                    )

        if span is md4c.SpanType.A:
            if self._current_aspan_ref_target:  # referenced link
                self.current_msgid += f'[{self._current_aspan_text}]'
                if self._current_aspan_ref_target != self._current_aspan_text:
                    self.current_msgid += (
                        f'[{self._current_aspan_ref_target}]'
                    )
                self._current_aspan_ref_target = None
            else:
                title = details['title'][0][1] if details['title'] else ''
                if self._current_aspan_text == details['href'][0][1]:
                    # autolink vs link clash (see implementation notes)
                    self.current_msgid += f'<{self._current_aspan_text}'
                    if title:
                        self.current_msgid += f' "{polib.escape(title)}"'
                    self.current_msgid += '>'
                else:
                    title_part = f' "{polib.escape(title)}"' if title else ''
                    href = details['href'][0][1]
                    self.current_msgid += (
                        f'[{self._current_aspan_text}]({href}{title_part})'
                    )
            self._inside_aspan = False
            self._current_aspan_text = ''
        elif span is md4c.SpanType.CODE:
            self._inside_codespan = False
            self._codespan_start_index = None

            # add backticks at the end for escape internal backticks
            if self._inside_aspan:
                self._current_aspan_text += (
                    self._codespan_backticks * self.code_end_string
                )
            else:
                self.current_msgid += (
                    self._codespan_backticks * self.code_end_string
                )
            self._codespan_backticks = None
        elif span is md4c.SpanType.IMG:
            referenced_target, imgspan_title = (None, None)
            imgspan_src = details['src'][0][1]
            if details['title']:
                imgspan_title = details['title'][0][1]
                for target, href, title in self.link_references:
                    if href == imgspan_src and title == imgspan_title:
                        referenced_target = target
                        break
            else:
                for target, href, _ in self.link_references:
                    if href == imgspan_src:
                        referenced_target = target
                        break

            alt_text = self._current_imgspan['text']
            img_markup = f'![{alt_text}]'
            if referenced_target:
                img_markup += f'[{referenced_target}]'
            else:
                img_markup += f'({imgspan_src}'
                if imgspan_title:
                    img_markup += f' "{polib.escape(imgspan_title)}"'
                img_markup += ')'

            self._current_imgspan = {}

            if self._inside_aspan:
                self._current_aspan_text += img_markup
            else:
                self.current_msgid += img_markup
        elif span is md4c.SpanType.U:
            self._inside_uspan = False
        elif span is md4c.SpanType.LATEXMATH_DISPLAY:
            self._inside_latexmath_display = False
            self.current_msgid += self._inside_latexmath_display_text
            self._inside_latexmath_display_text = ''
            self._save_current_msgid()

    def text(self, block, text):
        # raise 'text' event
        if raise_skip_event(self.events, 'text', self, block, text):
            return

        if not self._inside_htmlblock:
            if not self._inside_codeblock:
                if any([  # softbreaks
                    self._inside_liblock, self._inside_aspan,
                ]) and text == '\n':
                    text = ' '
                if not self.plaintext:
                    if self._current_imgspan:
                        self._current_imgspan['text'] = text
                        return
                    if self._inside_codespan:
                        # fix backticks for codespan start and end to escape
                        # internal backticks
                        self._codespan_backticks = min_not_max_chars_in_a_row(
                            self.code_start_string,
                            text,
                        ) - 1
                        backticks = (
                            self._codespan_backticks * self.code_start_string
                        )
                        self.current_msgid = (
                            f'{self.current_msgid[:self._codespan_start_index]}'
                            f'{backticks}'
                            f'{self.current_msgid[self._codespan_start_index:]}'
                        )
                        if self._inside_aspan:
                            self._current_aspan_text += text
                            return
                    elif self._inside_aspan:
                        self._current_aspan_text += text
                        return
                    elif text == self.italic_start_string:
                        text = self.italic_start_string_escaped
                    elif text == self.code_start_string:
                        text = self.code_start_string_escaped
                    elif text == self.italic_end_string:   # pragma: no cover
                        text = self.italic_end_string_escaped
                    elif text == self.code_end_string:   # pragma: no cover
                        text = self.code_end_string_escaped
                if self._inside_pblock:
                    text = text.replace('\n', ' ')
                if self._current_wikilink_target:
                    if text != self._current_wikilink_target:
                        # not self-referenced wikilink
                        self._current_wikilink_target = (
                            f'{self._current_wikilink_target}|{text}'
                        )
                    return
                if self._inside_latexmath_display:
                    self._inside_latexmath_display_text += text.strip()
                    return
                self.current_msgid += text
            elif not self.disable_next_codeblock and (
                self.include_codeblocks or self.include_next_codeblock
            ):
                self.current_msgid += text
        else:
            self._process_command(text)

    def _dump_link_references(self):
        if self.link_references:
            self.disable_next_block = False
            self.disable = False

            # 'link_reference' event
            pre_events = self.events.get('link_reference')

            for target, href, title in self.link_references:
                if pre_events:
                    skip = False
                    for event in pre_events:
                        if event(self, target, href, title) is False:
                            skip = True
                    if skip:
                        continue

                self.current_msgid = '[{}]:{}{}'.format(
                    target,
                    f' {href}' if href else '',
                    f' "{title}"' if title else '',
                )
                self._save_current_msgid(
                    msgstr=self.current_msgid,
                    fuzzy=True,
                )

    def extract(
        self,
        po_filepath=None,
        save=False,
        mo_filepath=None,
        po_encoding=None,
        md_encoding='utf-8',
        wrapwidth=78,
    ):
        if not po_filepath:
            self.po_filepath = ''

            if save:
                if os.environ.get('_MDPO_RUNNING') == 'true':
                    save_arg = '-s/--save'
                    po_filepath_arg = '-p/--po-filepath'
                else:
                    save_arg, po_filepath_arg = ('save', 'po_filepath')
                raise ValueError(
                    f"The argument '{save_arg}' does not make sense"
                    f" without passing the argument '{po_filepath_arg}'.",
                )
        else:
            self.po_filepath = po_filepath
            if not os.path.exists(po_filepath):
                self.po_filepath = ''

        pofile_kwargs = (
            {'autodetect_encoding': False, 'encoding': po_encoding}
            if po_encoding else {}
        )
        self.pofile = polib.pofile(
            self.po_filepath,
            wrapwidth=parse_wrapwidth_argument(wrapwidth),
            **pofile_kwargs,
        )
        for entry in self.pofile:
            if entry.obsolete:
                self.obsoletes = True
                break

        parser = md4c.GenericParser(
            0,
            **{ext: True for ext in self.extensions},
        )

        def _parse(content):
            parser.parse(
                content,
                self.enter_block,
                self.leave_block,
                (
                    self.enter_span if self.plaintext
                    else self.not_plaintext_enter_span
                ),
                (
                    self.leave_span if self.plaintext
                    else self.not_plaintext_leave_span
                ),
                self.text,
            )
            self._dump_link_references()

        if hasattr(self, 'content'):
            _parse(self.content)
        else:
            for filepath in self.filepaths:
                with open(filepath, encoding=md_encoding) as f:
                    self.content = f.read()
                self._current_markdown_filepath = filepath
                _parse(self.content)

                # reset state
                self.disable_next_block = False
                self.disable = False
                self.enable_next_block = False
                self.include_next_codeblock = False
                self.disable_next_codeblock = False
                self.link_references = None
                self._current_top_level_block_number = 0
                self._current_top_level_block_type = None

        if not self.preserve_not_found:
            remove_not_found_entries(
                self.pofile,
                self.found_entries,
            )
        elif self.mark_not_found_as_obsolete:
            mark_not_found_entries_as_obsoletes(
                self.pofile,
                self.found_entries,
            )
            self.obsoletes = True

        if self.metadata:
            self.pofile.metadata.update(self.metadata)

        if save and po_filepath:
            if self._saved_files_changed is False:
                self._saved_files_changed = save_file_checking_file_changed(
                    po_filepath,
                    str(self.pofile),
                    encoding=self.pofile.encoding,
                )
            else:
                self.pofile.save(fpath=po_filepath)
        if mo_filepath:
            self.pofile.save_as_mofile(mo_filepath)
        return self.pofile


def markdown_to_pofile(
    files_or_content,
    ignore=frozenset(),
    msgstr='',
    po_filepath=None,
    save=False,
    mo_filepath=None,
    plaintext=False,
    wrapwidth=78,
    mark_not_found_as_obsolete=True,
    preserve_not_found=True,
    location=True,
    extensions=DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS,
    po_encoding=None,
    md_encoding='utf-8',
    xheader=False,
    include_codeblocks=False,
    ignore_msgids=frozenset(),
    command_aliases=None,
    metadata=None,
    events=None,
    debug=False,
    **kwargs,
):
    """Extract all the msgids from Markdown content or files.

    Args:
        files_or_content (str, list): Glob path to Markdown files, a list of
            files or a string with Markdown content.
        ignore (list): Paths of files to ignore. Useful when a glob does not
            fit your requirements indicating the files to extract content.
            Also, filename or a dirname can be defined without indicate the
            full path.
        msgstr (str): Default message string for extracted msgids.
        po_filepath (str): File that will be used as :class:`polib.POFile`
            instance where to dump the new msgids and that will be used
            as source checking not found strings that will be marked as
            obsolete if is the case (see ``save`` and
            ``mark_not_found_as_obsolete`` optional parameters).
        save (bool): Save the new content to the PO file indicated in the
            parameter ``po_filepath``. If is enabled and ``po_filepath`` is
            ``None`` a ``ValueError`` will be raised.
        mo_filepath (str): The resulting PO file will be compiled to a MO file
            and saved in the path specified at this parameter.
        plaintext (bool): If you pass ``True`` to this parameter (as default)
                the content will be extracted as is, without markup characters
                included.
                Passing ``plaintext`` as ``False``, extracted msgids
                will contain some markup characters used to appoint the
                location of ```inline code```, ``**bold text**``,
                ``*italic text*`` and ```[links]```, that might be useful
                for you. It depends on the use you are going to give to
                this library activate this mode (``plaintext=False``) or not.
        wrapwidth (int): Wrap width for po file indicated at ``po_filepath``
            parameter. If negative, 0, 'inf' or 'math.inf' the content won't
            be wrapped.
        mark_not_found_as_obsolete (bool): The strings extracted from markdown
            that will not be found inside the provided PO file will be marked
            as obsolete.
        preserve_not_found (bool): The strings extracted from markdown that
            will not be found inside the provided PO file wouldn't be removed.
            Only has effect if ``mark_not_found_as_obsolete`` is ``False``.
        location (bool): Store references of top-level blocks in which are
            found the messages in PO file `#: reference` comments.
        extensions (list): md4c extensions used to parse markdown content,
            formatted as a list of 'pymd4c' keyword arguments. You can see all
            available at `pymd4c documentation <https://pymd4c.dcpx.org
            /api.html#option-flags>`_.
        po_encoding (str): Resulting PO file encoding.
        md_encoding (str): Markdown content encoding.
        xheader (bool): Indicates if the resulting PO file will have the mdpo
            x-header included.
        include_codeblocks (bool): Include all code blocks found inside PO file
            result. This is useful if you want to translate all your blocks
            of code. Equivalent to append ``<!-- mdpo-include-codeblock -->``
            command before each code block.
        ignore_msgids (list): List of msgids ot ignore from being extracted.
        command_aliases (dict): Mapping of aliases to use custom mdpo command
            names in comments. The ``mdpo-`` prefix in command names resolution
            is optional. For example, if you want to use ``<!-- mdpo-on -->``
            instead of ``<!-- mdpo-enable -->``, you can pass the dictionaries
            ``{"mdpo-on": "mdpo-enable"}`` or ``{"mdpo-on": "enable"}`` to this
            parameter.
        metadata (dict): Metadata to include in the produced PO file. If the
            file contains previous metadata fields, these will be updated
            preserving the values of the already defined.
        events (dict): Preprocessing events executed during the parsing
            process that can be used to customize the extraction process.
            Takes functions or list of functions as values. If one of these
            functions returns ``False``, that part of the parsing is skipped
            by ``md2po``. Available events are the next:

            * ``enter_block(self, block, details)``: Executed when the parsing
              a Markdown block starts.
            * ``leave_block(self, block, details)``: Executed when the parsing
              a Markdown block ends.
            * ``enter_span(self, span, details)``: Executed when the parsing of
              a Markdown span starts.
            * ``leave_span(self, span, details)``: Executed when the parsing of
              a Markdown span ends.
            * ``text(self, block, text)``: Executed when the parsing of text
              starts.
            * ``command(self, mdpo_command, comment, original command)``:
              Executed when a mdpo HTML command is found.
            * ``msgid(self, msgid, msgstr, msgctxt, tcomment, flags)``:
              Executed when a msgid is going to be stored.
            * ``link_reference(self, target, href, title)``: Executed when a
              link reference is going to be stored.

            You can also define the location of these functions by strings
            with the syntax ``path/to/file.py::function_name``.

            All ``self`` arguments are an instance of Md2Po parser. You can
            take advanced control of the parsing process manipulating the
            state of the parser. For example, if you want to skip a certain
            msgid to be included, you can do:

            .. code-block:: python

               def msgid_event(self, msgid, *args):
                   if msgid == 'foo':
                       self.disable_next_block = True
        debug (bool): Add events displaying all parsed elements in the
            extraction process.
        **kwargs: Extra arguments passed to
            :py:class:`mdpo.md2po.Md2Po` constructor.

    Examples:
        >>> content = 'Some text with `inline code`'
        >>> entries = markdown_to_pofile(content, plaintext=True)
        >>> {e.msgid: e.msgstr for e in entries}
        {'Some text with inline code': ''}
        >>> entries = markdown_to_pofile(content)
        >>> {e.msgid: e.msgstr for e in entries}
        {'Some text with `inline code`': ''}
        >>> entries = markdown_to_pofile(content, msgstr='Default message')
        >>> {e.msgid: e.msgstr for e in entries}
        {'Some text with `inline code`': 'Default message'}

    Returns:
        :class:`polib.POFile` Pofile instance with new msgids included.
    """
    return Md2Po(
        files_or_content,
        ignore=ignore,
        msgstr=msgstr,
        plaintext=plaintext,
        mark_not_found_as_obsolete=mark_not_found_as_obsolete,
        preserve_not_found=preserve_not_found,
        location=location,
        extensions=extensions,
        xheader=xheader,
        include_codeblocks=include_codeblocks,
        ignore_msgids=ignore_msgids,
        command_aliases=command_aliases,
        metadata=metadata,
        events=events,
        debug=debug,
        **kwargs,
    ).extract(
        po_filepath=po_filepath,
        save=save,
        mo_filepath=mo_filepath,
        po_encoding=po_encoding,
        md_encoding=md_encoding,
        wrapwidth=wrapwidth,
    )



"""Markdown related utilities for mdpo."""


LINK_REFERENCE_REGEX = (
    r'^\[([^\]]+)\]:\s+<?([^\s>]+)>?\s*["\'\(]?([^"\'\)]+)?'
)


def parse_link_references(content):
    """Parses link references found in a Markdown content.

    Args:
        content (str): Markdown content to be parsed.

    Returns:
        list: Tuples with 3 values, target, href and title for each link
        reference.
    """
    import re

    link_reference_re = re.compile(LINK_REFERENCE_REGEX)

    response = []
    for line in content.splitlines():
        linestrip = line.strip()
        if linestrip and linestrip.startswith('['):
            match = re.search(link_reference_re, linestrip)
            if match:
                response.append(match.groups())
    return response


def solve_link_reference_targets(translations):
    """Solve link reference targets in markdown blocks.

    Given a dictionary of msgid/msgstr translations, those link references
    targets will be resolved and returned in a new dictionary.

    Args:
        translations (dict): Mapping of msgid-msgstr entries from which
            the resolved translations will be extracted.

    Returns:
        dict: New created messages with solved link reference targets.
    """
    import re

    link_refereced_link_re = re.compile(r'\[([^\]]+)\]\[([^\]\s]+)\]')
    link_reference_re = re.compile(LINK_REFERENCE_REGEX)

    solutions = {}

    # dictionary with defined link references and their targets
    link_references_text_targets = []

    # compound by dictionaries with `original_msgid`, `original_msgstr` and
    # `link_reference_matchs`
    msgid_msgstrs_with_links = []

    # discover link reference definitions
    for msgid, msgstr in translations.items():
        if msgid.startswith('['):  # filter for performance improvement
            msgid_match = re.search(link_reference_re, msgid.lstrip(' '))
            if msgid_match:
                msgstr_match = re.search(link_reference_re, msgstr.lstrip(' '))
                if msgstr_match:
                    link_references_text_targets.append((
                        msgid_match.groups(),
                        msgstr_match.groups(),
                    ))
        msgid_matchs = re.findall(link_refereced_link_re, msgid)
        if msgid_matchs:
            msgstr_matchs = re.findall(link_refereced_link_re, msgstr)
            if msgstr_matchs:
                msgid_msgstrs_with_links.append((
                    msgid, msgstr, msgid_matchs, msgstr_matchs,
                ))

    # original msgid, original msgstr,
    # msgid link reference matchs, msgstr link reference matchs
    for (
            orig_msgid, orig_msgstr, msgid_linkr_groups, msgstr_linkr_groups,
    ) in msgid_msgstrs_with_links:
        # search if msgid and link reference matchs are inside
        # `link_references_text_targets`
        #
        # if so, replace in original messages link referenced targets with
        # real targets and store them in solutions
        new_msgid, new_msgstr = (None, None)

        for msgid_linkr_group in msgid_linkr_groups:
            for link_reference_text_targets in link_references_text_targets:

                if link_reference_text_targets[0][0] == msgid_linkr_group[1]:
                    replacer = (
                        f'[{msgid_linkr_group[0]}][{msgid_linkr_group[1]}]'
                    )
                    replacement = (
                        f'[{msgid_linkr_group[0]}]'
                        f'({link_reference_text_targets[0][1]})'
                    )

                    if new_msgid is None:
                        # first referenced link replacement in msgid
                        new_msgid = orig_msgid.replace(replacer, replacement)
                    else:
                        # consecutive referenced link replacements in msgid
                        new_msgid = new_msgid.replace(replacer, replacement)
                    break

        # the same game as above, but now for msgstrs

        for msgstr_linkr_group in msgstr_linkr_groups:
            for link_reference_text_targets in link_references_text_targets:

                if link_reference_text_targets[1][0] == msgstr_linkr_group[1]:
                    replacer = (
                        f'[{msgstr_linkr_group[0]}][{msgstr_linkr_group[1]}]'
                    )
                    replacement = (
                        f'[{msgstr_linkr_group[0]}]'
                        f'({link_reference_text_targets[1][1]})'
                    )

                    if new_msgstr is None:
                        # first referenced link replacement in msgid
                        new_msgstr = orig_msgstr.replace(replacer, replacement)
                    else:
                        # consecutive referenced link replacements in msgid
                        new_msgstr = new_msgstr.replace(replacer, replacement)
                    break

        # store in solutions
        solutions[new_msgid] = new_msgstr

    return solutions



"""Workaround for polib issue and required improvement to make work mdpo.

See https://github.com/izimobil/polib/pulls for details.
"""


def poentry__cmp__(
    self,
    other,
    compare_obsolete=True,
    compare_msgstr=True,
    compare_occurrences=True,
):
    """Compare entries like ``__cmp__`` method of :py:class:`polib.POEntry`.

    This function add custom entries comparison capabilities to
    :py:class:`polib.POEntry`, needed by mdpo to prevent some errors.

    Args:
        self (:py:class:`polib.POEntry`): Entry to be compared.
        other (:py:class:`polib.POEntry`): Entry to compare against.
        compare_obsolete (bool): Indicates if the ``obsolete`` property of the
            entries will be used comparing them.
        compare_msgstr (bool): Indicates if the ``msgstr`` property of the
            entries will be used comparing them.
        compare_occurrences (bool): Indicates if the ``occurrences`` property
            of the entries will be used comparing them.
    """
    if compare_obsolete and self.obsolete != other.obsolete:
        return -1 if self.obsolete else 1
    if compare_occurrences:
        # `.copy` is an equivalent for full slice [:], but is faster
        # `.copy` doesn't exists for lists in Python2 so for this reason,
        # this change has not been introduced in polib
        occ1 = sorted(self.occurrences.copy())
        occ2 = sorted(other.occurrences.copy())
        if occ1 > occ2:
            return 1
        if occ1 < occ2:
            return -1
    msgctxt = self.msgctxt or '0'
    othermsgctxt = other.msgctxt or '0'
    if msgctxt > othermsgctxt:
        return 1
    if msgctxt < othermsgctxt:
        return -1
    msgid_plural = self.msgid_plural or '0'
    othermsgid_plural = other.msgid_plural or '0'
    if msgid_plural > othermsgid_plural:
        return 1
    if msgid_plural < othermsgid_plural:
        return -1
    msgstr_plural = self.msgstr_plural or '0'
    othermsgstr_plural = other.msgstr_plural or '0'
    if msgstr_plural > othermsgstr_plural:
        return 1
    if msgstr_plural < othermsgstr_plural:
        return -1
    if self.msgid > other.msgid:
        return 1
    if self.msgid < other.msgid:
        return -1
    if compare_msgstr:
        if self.msgstr > other.msgstr:
            return 1
        if self.msgstr < other.msgstr:
            return -1
    return 0



"""mdpo command line interface utilities."""

import argparse
import sys

from importlib_metadata_argparse_version import ImportlibMetadataVersionAction

from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS
from mdpo.text import and_join, parse_escaped_pairs


SPHINX_IS_RUNNING = 'sphinx' in sys.modules
OPEN_QUOTE_CHAR = '”' if SPHINX_IS_RUNNING else '"'
CLOSE_QUOTE_CHAR = '”' if SPHINX_IS_RUNNING else '"'


def cli_codespan(value, cli=True):
    """Command line codespan wrapper.

    This is a compatibility function to make CLI codespans looks good in
    sphinx-argparse-cli documentation and using ``--help`` option in CLI
    itself. sphinx-argparse-cli expects the usage of double backticks for
    codespans, but that format is ugly in CLI.

    Args:
        value (str): Value to wrap.
        cli (bool): Wrap when used from command line.
    """
    if SPHINX_IS_RUNNING:
        return f'``{value}``'
    if cli:  # pragma: no cover
        return f"'{value}'"
    return value


def parse_escaped_pairs_cli_argument(
    pairs,
    value_error_message,
    key_error_message,
):
    """Parse a key argument made by key-value pairs.

    If an error happens, shows an appropiate message and exists with code 1.

    Args:
        pairs (list): List of key-value pairs.
        value_error_message (str): Error message schema shown when a pair
            can't be parsed.
        key_error_message (str): Error message schema shown when a key is
            repeated.

    Returns:
        dict: Parsed key-value pairs.
    """
    try:
        return parse_escaped_pairs(pairs)
    except ValueError as err:
        sys.stderr.write(value_error_message.format(err.args[0]))
        sys.exit(1)
    except KeyError as err:
        sys.stderr.write(key_error_message.format(err.args[0]))
        sys.exit(1)


def parse_command_aliases_cli_arguments(command_aliases):
    """Parse ``--command-alias`` argument values passed to CLIs.

    If a value can't be passed or a custom command is duplicated, writes an
    appropiate error message to STDERR and exits with code 1.

    Args:
        command_aliases (list): Values taken by ``--command-alias`` arguments.

    Returns:
        dict: Command aliases mapping ni the format accepted by the API.
    """
    return parse_escaped_pairs_cli_argument(
        command_aliases,
        (
            "The value '{}' passed to argument --command-alias"
            " can't be parsed. Please, separate the pair "
            "'<custom-command:mdpo-command>' with a ':' character.\n"
        ),
        (
            "Multiple resolutions for '{}' alias passed to"
            ' --command-alias arguments.\n'
        ),
    )


def parse_metadata_cli_arguments(metadata):
    """Parse ``--metadata`` argument values passed to CLIs.

    If a value can't be passed or a metadata key is duplicated, writes an
    appropiate error message to STDERR and exits with code 1.

    Args:
        metadata (list): Values taken by ``--metadata`` arguments.

    Returns:
        dict: Metadata mapping ni the format accepted by the API.
    """
    return parse_escaped_pairs_cli_argument(
        metadata,
        (
            "The value '{}' passed to argument --metadata"
            " can't be parsed. Please, separate the pair "
            "'<key:value>' with a ':' character.\n"
        ),
        (
            "Repeated key '{}' passed to --metadata arguments.\n"
        ),
    )


def add_common_cli_first_arguments(parser):
    """Add common mdpo arguments to an argument parser at the beginning.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '-h', '--help', action='help',
        help='Show this help message and exit.',
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        '-v', '--version', action=ImportlibMetadataVersionAction,
        version_from='mdpo',
        version='%(prog)s %(version)s',
        help='Show program version number and exit.',
    )
    parser.add_argument(
        '-q', '--quiet', action='store_true',
        help='Do not print output to STDOUT.',
    )


def add_command_alias_argument(parser):
    """Add the ``--command-alias`` argument to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    mdpo_on_mdpo_enable = cli_codespan(
        f'--command-alias {OPEN_QUOTE_CHAR}mdpo-on:mdpo-enable'
        f'{CLOSE_QUOTE_CHAR}',
    )
    mdpo_on_enable = cli_codespan(
        f'--command-alias {OPEN_QUOTE_CHAR}mdpo-on:enable'
        f'{CLOSE_QUOTE_CHAR}',
    )
    command_alias_help_example = (
        ' For example, if you want to use "<!-- mdpo-on -->" instead of'
        f' "<!-- mdpo-enable -->", you can pass either {mdpo_on_mdpo_enable}'
        f' or {mdpo_on_enable} arguments.'
    )
    parser.add_argument(
        '--command-alias', dest='command_aliases', default=[], action='append',
        metavar='CUSTOM-COMMAND:MDPO-COMMAND',
        help='Aliases to use custom mdpo command names in comments. This'
             ' argument can be passed multiple times in the form'
             ' "<custom-command>:<mdpo-command>". The \'mdpo-\' prefix in'
             ' command names resolution is optional.'
             f'{command_alias_help_example}',
    )


def add_extensions_argument(parser):
    """Add the ``-x/--extension`` argument to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '-x', '--extension', '--ext', dest='extensions', action='append',
        default=None,
        help='md4c extension used to parse markdown content formatted as'
             ' pymd4c extension keyword arguments. This argument can be passed'
             ' multiple times. If it is not passed, next extensions are used:'
             f' {and_join(DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS)}.'
             ' You can see all available at'
             ' https://pymd4c.dcpx.org/api.html#parser-option-flags',
        metavar='EXTENSION',
    )


def add_debug_option(parser):
    """Add the ``-D/--debug`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '-D', '--debug', dest='debug', action='store_true',
        help='Print useful messages in the parsing process showing the'
             ' contents of all Markdown elements.',
    )


def add_nolocation_option(parser):
    """Add the ``--no-location/--nolocation`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '--no-location', '--nolocation', dest='location', action='store_false',
        help="Do not write '#: filename:line' lines. Note that using this"
             ' option makes it harder for technically skilled translators to'
             ' understand the context of each message. Same as'
             f' {cli_codespan("gettext --no-location")}.',
    )


def add_encoding_arguments(
    parser,
    po_encoding_help=None,
    markup_encoding='md',
):
    """Add ``--po-encoding`` ``--md-encoding`` arguments to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
        po_encoding_help (str): Help text for the ``--po-encoding`` argument.
        markup_encoding (str): Type of markup encoding, can be ``"md"`` or
            ``"html"``.
    """
    parser.add_argument(
        f'--{markup_encoding}-encoding',
        dest=f'{markup_encoding}_encoding',
        default='utf-8',
        help=f'{"Markdown" if markup_encoding == "md" else "HTML"}'
             ' content encoding.',
        metavar='ENCODING',
    )

    po_encoding_help = (
        'PO files encoding. If you need different encodings for each'
        ' file, you must define them in the "Content-Type" field of each'
        " PO file metadata, in the form 'Content-Type: text/plain;"
        " charset=<ENCODING>'."
    ) if po_encoding_help is None else po_encoding_help
    parser.add_argument(
        '--po-encoding', dest='po_encoding', default=None, metavar='ENCODING',
        help=po_encoding_help,
    )


def add_wrapwidth_argument(
    parser,
    markup='po',
    markup_prefix=False,
    short=True,
    default='78',
    help_to_render=None,
):
    """Add a ``--wrapwidth`` argument to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
        markup (str): For which type of files the argument will affect. Either
            ``"md"`` for Markdown or ``"po"`` for PO files.
        markup_prefix (bool): Add the ``markup`` prefix in the argument using
            the format ``--<markup>-wrapwidth``.
        short (bool): Add the short version of the argument (``-w``).
        default (str): Default value.
        help_to_render (str): String used to indicate the content that will be
            wrapped according to the argument.
    """
    args = [f'--{markup}-wrapwidth' if markup_prefix else '--wrapwidth']
    if short:
        args.append('-w')

    kwargs = {
        'metavar': 'N/inf',
        'type': str,
        'default': default,
    }
    if help_to_render is not None:
        to_render = help_to_render
    elif markup == 'po':
        to_render = (
            'the PO file indicated at parameter'
            f' {cli_codespan("--po-filepath")}'
        )
    else:
        to_render = 'the Markdown output, when possible'
    kwargs['help'] = (
        f"Maximum width rendering {to_render}. If negative, '0' or 'inf',"
        ' the content will not be wrapped.'
    )
    parser.add_argument(*args, **kwargs)


def add_check_option(parser):
    """Add the ``--check`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '--check', dest='check_saved_files_changed', action='store_true',
        help='Run in check mode, which returns code 2 at exit when a file'
             ' has been changed or previously did not exist.',
    )


def add_event_argument(parser):
    """Add the ``--event`` optional argument to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '-e', '--event', dest='events', default=[], action='append',
        metavar='event_name: path/to/file.py::function_name',
        help='Custom events executed during the parser. They are used for'
             ' customize the output. See the documentation for available'
             ' event names. This argument can be passed multiple times.',
    )


def add_include_codeblocks_option(parser):
    """Add the ``--include-codeblocks`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '-c', '--include-codeblocks',
        dest='include_codeblocks', action='store_true',
        help='Include all code blocks found inside PO file result. This is'
             ' useful if you want to translate all your blocks of code.'
             ' Equivalent to append'
             f' {cli_codespan("<!-- mdpo-include-codeblock -->")} command'
             ' before each code block.',
    )


def add_no_obsolete_option(parser):
    """Add the ``--no-obsolete`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '--no-obsolete', dest='no_obsolete', action='store_true',
        help=(
            'If obsolete entries found in PO files exit with non'
            ' zero code.'
        ),
    )


def add_no_empty_msgstr_option(parser):
    """Add the ``--no-empty-msgstr`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Parser to extend.
    """
    parser.add_argument(
        '--no-empty-msgstr', dest='no_empty_msgstr', action='store_true',
        help=(
            'If empty msgstrs found in PO files exit with non zero code.'
        ),
    )


def parse_event_argument(value):
    """Parse ``--event`` CLI argument values.

    Args:
        value (list): Event names and function paths in the form
            ``event_name: path/to/file.py::func``.

    Returns:
        dict: Mapping of event names and `file::function` paths.
    """
    events = {}
    for event_name_filefunc in value:
        event_name, filefunc = event_name_filefunc.split(':', maxsplit=1)
        events[event_name.strip()] = filefunc.strip()
    return events



"""mdpo package."""


func_package_map = {
    'markdown_pofile_to_html': 'mdpo2html',
    'markdown_to_pofile': 'md2po',
    'markdown_to_pofile_to_markdown': 'md2po2md',
    'pofile_to_markdown': 'po2md',
}

__all__ = list(func_package_map)  # noqa: PLE0605 (not implemented by Ruff)


def __getattr__(name):
    """Implement PEP 562 to avoid uneeded imports in CLIs."""
    import importlib

    try:
        return getattr(
            importlib.import_module(f'mdpo.{func_package_map[name]}'),
            name,
        )
    except KeyError:
        raise ImportError(
            f"cannot import name '{name}' from 'mdpo' ({__file__})",
            name=name,
            path='mdpo',
        ) from None



"""HTML-produced-from-Markdown files translator using PO files as reference."""

import contextlib
import html
import re
import warnings
from collections import OrderedDict
from html.parser import HTMLParser

import md4c
import polib

from mdpo.command import (
    normalize_mdpo_command_aliases,
    parse_mdpo_html_command,
)
from mdpo.io import save_file_checking_file_changed, to_file_content_if_is_file
from mdpo.md import solve_link_reference_targets
from mdpo.po import (
    paths_or_globs_to_unique_pofiles,
    pofiles_to_unique_translations_dicts,
)


PROCESS_REPLACER_TAGS = [
    'p', 'li', 'h1', 'h2', 'h3',
    'h4', 'h5', 'h6', 'td', 'th',
]
ALIGNMENT_CHARS = ['\n', ' ', '\t', '\r']


def _html_attrs_to_str(attrs):
    """Converts a dictionary of HTML attributes to its HTML representation.

    Args:
        attrs (dict): Attributes with values

    Returns:
        str: HTML attributes string ready to be used inside a HTML tag.
    """
    response = ''
    for name, value in attrs.items():
        response += name
        if value is not None:
            response += f'="{value}"'
        response += ' '
    return response.rstrip(' ')


class MdPo2HTML(HTMLParser):
    def __init__(
        self,
        pofiles,
        ignore=frozenset(),
        merge_adjacent_markups=False,
        code_tags=frozenset(['code']),
        bold_tags=frozenset(['b', 'strong']),
        italic_tags=frozenset(['em', 'i']),
        link_tags=frozenset(['a']),
        image_tags=frozenset(['img']),
        ignore_grouper_tags=frozenset(['div', 'hr']),
        po_encoding=None,
        command_aliases=None,
        _check_saved_files_changed=None,
    ):
        self.pofiles = paths_or_globs_to_unique_pofiles(
            pofiles,
            ignore,
            po_encoding=po_encoding,
        )
        self.output = ''
        self.replacer = []
        self._raw_replacement = ''
        self.context = []
        self.current_msgctxt = None

        self.translations = None
        self.translations_with_msgctxt = None

        self.disable = False
        self.disable_next_block = False
        self.enable_next_block = False
        self.disabled_entries = []

        # custom mdpo command resolution
        self.command_aliases = normalize_mdpo_command_aliases(
            command_aliases or {},
        )

        # lazy translators mode
        self.merge_adjacent_markups = merge_adjacent_markups

        # pre-commit mode
        self._saved_files_changed = (
            False if _check_saved_files_changed is not None else None
        )

        # references of msgids and their msgstrs changing referenced link
        # targets by their real target, needed to support link references
        # (this variable is redefined when the first link tag `a` is found)
        self.real_link_reference_targets = None

        # code markup
        self.code_tags = code_tags

        # bold markup
        self.bold_tags = bold_tags

        # italic markup
        self.italic_tags = italic_tags

        # link markup
        self.link_tags = link_tags

        # image markup
        self.image_tags = image_tags

        self.ignore_grouper_tags = ignore_grouper_tags

        self.markup_tags = []
        self.markup_tags.extend(self.code_tags)
        self.markup_tags.extend(self.bold_tags)
        self.markup_tags.extend(self.italic_tags)

        self.html_renderer = md4c.HTMLRenderer(md4c.MD_FLAG_TABLES)

        super().__init__()

    def _merge_adyacent_tags(self, html, template_tags):
        for tags_group in [self.bold_tags, self.italic_tags]:
            regexes = []
            for tag in tags_group:
                if tag not in template_tags:
                    continue
                for _tag in tags_group:
                    if _tag not in template_tags:
                        continue
                    regex = fr'</{tag}>\s*<{_tag}>'
                    if regex not in regexes:
                        regexes.append(regex)

            for regex in regexes:
                html = re.sub(regex, ' ', html)

        return html

    def _remove_lastline_from_output_if_empty(self):
        split_output = self.output.split('\n')
        if not split_output[-1]:
            self.output = '\n'.join(split_output)[:-1]

    def _process_replacer(self):

        template_tags = []
        raw_html_template, _current_replacement = ('', '')

        _current_link_target = ''

        _last_start_tag = None
        _last_end_tag = None
        _inside_code = False
        while self.replacer:
            handle, handled, attrs = self.replacer.pop(0)
            if handle == 'start':
                template_tags.append(handled)

                attrs_repr = (
                    f' {_html_attrs_to_str(attrs)}' if attrs else ''
                )

                if handled in self.code_tags:
                    _current_replacement += '`'
                    raw_html_template += f'<{handled}{attrs_repr}>'
                elif handled in self.bold_tags:
                    _current_replacement += '**'
                    raw_html_template += f'<{handled}{attrs_repr}>'
                elif handled in self.italic_tags:
                    _current_replacement += '*'
                    raw_html_template += f'<{handled}{attrs_repr}>'
                elif handled in self.link_tags:
                    title, href = attrs.get('title'), attrs['href']
                    _current_link_target += f'({href}'
                    if title:
                        _current_link_target += f' "{title}"'
                    _current_link_target += ')'

                    raw_html_template += f'<{handled}'

                    for attr in attrs:
                        if attr in ['title', 'href']:
                            raw_html_template += f' {attr}="{{}}"'
                        #     These attributes are not included in output
                    # if attrs_except_href_title:
                    # raw_html_template += _html_attrs_to_str(
                    #     attrs_except_href_title) + '>'
                    raw_html_template += '>'
                else:
                    raw_html_template += f'<{handled}{attrs_repr}>'
                _last_start_tag = handled
                if _last_start_tag == 'code':
                    _inside_code = True

            elif handle == 'data':
                if not _inside_code:
                    while '  ' in handled:
                        handled = handled.replace('  ', ' ')

                if all((ch in ALIGNMENT_CHARS) for ch in handled):
                    raw_html_template += handled
                    _current_replacement += handled
                else:
                    raw_html_template += '{}'
                    if _current_link_target:
                        _current_replacement += (
                            f'[{handled}]{_current_link_target}'
                        )
                        _current_link_target = ''
                    else:
                        _current_replacement += handled
            elif handle == 'end':
                raw_html_template += f'</{handled}>'
                if handled in self.code_tags:
                    _current_replacement += '`'
                elif handled in self.bold_tags:
                    _current_replacement += '**'
                elif handled in self.italic_tags:
                    _current_replacement += '*'
                _last_end_tag = handled
                if _last_end_tag == 'code':
                    _inside_code = False
            elif handle == 'comment':
                raw_html_template += f'<!--{handled}-->'
            elif handle == 'startend':
                if handled in self.image_tags:
                    alt, src = attrs.get('alt', ''), attrs['src']
                    _current_replacement += f'![{alt}]({src}'
                    title = attrs.get('title')
                    if title:
                        _current_replacement += f' "{title}"'
                    _current_replacement += ')'

                    raw_html_template += '{}'
                else:
                    attrs_repr = (
                        f' {_html_attrs_to_str(attrs)}' if attrs else ''
                    )
                    raw_html_template += f'<{handled}{attrs_repr}/>'

        _current_replacement = html.unescape(_current_replacement)

        if (self.disable and not self.enable_next_block) \
                or self.disable_next_block:
            replacement = _current_replacement

            self.disabled_entries.append(
                polib.POEntry(
                    msgid=replacement,
                    msgstr='',
                    msgctxt=self.current_msgctxt,
                ),
            )
        else:
            if self.current_msgctxt:
                replacement = self.translations_with_msgctxt[
                    self.current_msgctxt
                ].get(_current_replacement)
            else:
                replacement = self.translations.get(_current_replacement)
            if not replacement:
                replacement = _current_replacement

        for tt in template_tags:
            for tags_group in [self.bold_tags, self.italic_tags]:
                if tt in tags_group:
                    for tag in tags_group:
                        if tag not in template_tags:
                            template_tags.append(tag)

        html_before_first_replacement = raw_html_template.split('{')[0]
        html_after_last_replacement = raw_html_template.split('}')[-1]
        for tags_group in [self.bold_tags, self.italic_tags, self.code_tags]:
            for tag in tags_group:
                html_before_first_replacement = \
                    html_before_first_replacement.split(f'<{tag}>')[0]
                html_after_last_replacement = \
                    html_after_last_replacement.split(f'</{tag}>')[-1]
            html_before_first_replacement = \
                html_before_first_replacement.split('<a href="')[0]

        html_inner = '\n'.join(
            self.html_renderer.parse(replacement).split('\n')[:-1],
        )
        html_inner = '>'.join(html_inner.split('>')[1:])
        html_inner = '<'.join(html_inner.split('<')[:-1])

        html_template = html_before_first_replacement + html_inner + \
            html_after_last_replacement

        if self.merge_adjacent_markups:
            html_template = self._merge_adyacent_tags(
                html_template,
                template_tags,
            )

        self.output += html_template
        self.context = []

        self.disable_next_block = False
        self.enable_next_block = False
        self.current_msgctxt = None

    def handle_starttag(self, tag, attrs):

        if tag in self.ignore_grouper_tags:
            self.context.append(tag)
            attrs_repr = (
                f' {_html_attrs_to_str(OrderedDict(attrs))}' if attrs else ''
            )
            self.output += f'<{tag}{attrs_repr}>'
        elif (
            (self.context and self.context[0] in self.ignore_grouper_tags)
            or (tag == 'ul' and not self.context)
            or (tag in {'blockquote', 'table', 'thead', 'tbody', 'tr'})
        ):
            attrs_repr = (
                f' {_html_attrs_to_str(OrderedDict(attrs))}' if attrs else ''
            )
            self.output += f'<{tag}{attrs_repr}>'
        else:
            if tag == 'a' and self.real_link_reference_targets is None:
                # extend translations to populate link reference msgid-msgstrs
                # with real targets
                self.real_link_reference_targets = (
                    solve_link_reference_targets(self.translations)
                )
                self.translations.update(self.real_link_reference_targets)

            self.replacer.append(('start', tag, OrderedDict(attrs)))
            self.context.append(tag)

    def handle_endtag(self, tag):

        if tag in self.ignore_grouper_tags:
            self.output += f'</{tag}>'
            if self.context:
                self.context.pop()
        elif self.context and self.context[0] in self.ignore_grouper_tags:
            self.output += f'</{tag}>'
        elif tag in PROCESS_REPLACER_TAGS:
            self.replacer.append(('end', tag, None))
            self._process_replacer()
        elif tag in ['ul', 'blockquote', 'tr', 'table', 'thead', 'tbody']:
            self.output += f'</{tag}>'
        else:
            self.replacer.append(('end', tag, None))
            if self.context:
                self.context.pop()

    def handle_startendtag(self, tag, attrs):

        if not self.replacer:
            self.output += self.get_starttag_text()
        else:
            self.replacer.append(('startend', tag, OrderedDict(attrs)))

    def handle_data(self, data):

        if data:
            if (
                not self.replacer
                or (
                    self.context
                    and self.context[0] in self.ignore_grouper_tags
                )
                or not self.context
            ):
                self.output += data
            else:
                if self.context:
                    data = data.replace('\n', ' ')
                self.replacer.append(('data', data, None))

    def handle_comment(self, data):

        if self.replacer:
            self.replacer.append(('comment', data, None))
        else:
            data_as_comment = f'<!--{data}-->'
            command, comment = parse_mdpo_html_command(
                data_as_comment,
            )
            if command is None:
                self.output += data_as_comment
            else:
                self._remove_lastline_from_output_if_empty()

                with contextlib.suppress(KeyError):  # not custom command
                    command = self.command_aliases[command]

                if command in (
                    'mdpo-disable-next-block',
                    'mdpo-disable-next-line',
                ):
                    self.disable_next_block = True
                elif command == 'mdpo-disable':
                    self.disable = True
                elif command == 'mdpo-enable':
                    self.disable = False
                elif command in (
                    'mdpo-enable-next-block',
                    'mdpo-enable-next-line',
                ):
                    self.enable_next_block = True
                elif command == 'mdpo-context' and comment:
                    self.current_msgctxt = comment
                elif command == 'mdpo-include-codeblock':
                    warnings.warn(
                        'Code blocks translations are not supported'
                        ' by mdpo2html implementation.',
                        SyntaxWarning,
                        stacklevel=2,
                    )
                else:
                    self.output += data_as_comment

    def translate(self, filepath_or_content, save=None, html_encoding='utf-8'):
        content = to_file_content_if_is_file(
            filepath_or_content,
            encoding=html_encoding,
        )

        self.translations, self.translations_with_msgctxt = (
            pofiles_to_unique_translations_dicts(self.pofiles)
        )

        self.feed(content)

        if save:
            if self._saved_files_changed is False:
                self._saved_files_changed = save_file_checking_file_changed(
                    save,
                    self.output,
                    encoding=html_encoding,
                )
            else:
                with open(save, 'w', encoding=html_encoding) as f:
                    f.write(self.output)

        self.reset()

        return self.output


def markdown_pofile_to_html(
    filepath_or_content,
    pofiles,
    ignore=frozenset(),
    save=None,
    po_encoding=None,
    html_encoding='utf-8',
    command_aliases=None,
    **kwargs,
):
    r"""HTML-produced-from-Markdown file translator using PO files.

    Produces a translated HTML file given a previous HTML file (created by a
    Markdown-to-HTML processor) and a set of PO files as reference for msgstrs.

    Args:
        filepath_or_content (str): HTML whose content wants to be translated.
        pofiles (str): Glob for set of pofiles used as reference translating
            the strings to another language.
        ignore (list): List of paths to pofiles to ignore, useful if the glob
            patterns in ``pofiles`` parameter does not fit your requirements.
        save (str): If you pass this parameter as a path to one HTML file,
            even if does not exists, will be saved in the path the output of
            the function.
        html_encoding (str): HTML content encoding.
        po_encoding (str): PO files encoding. If you need different encodings
            for each file, you must define it in the "Content-Type" field of
            each PO file metadata, in the form
            ``"Content-Type: text/plain; charset=<ENCODING>\n"``.
        command_aliases (dict): Mapping of aliases to use custom mdpo command
            names in comments. The ``mdpo-`` prefix in command names resolution
            is optional. For example, if you want to use ``<!-- mdpo-on -->``
            instead of ``<!-- mdpo-enable -->``, you can pass the dictionaries
            ``{"mdpo-on": "mdpo-enable"}`` or ``{"mdpo-on": "enable"}`` to this
            parameter.
        **kwargs: Extra keyword arguments passed to
            :py:class:`mdpo.mdpo2html.MdPo2HTML` constructor.

    .. rubric:: Known limitations:

    * This implementation doesn't include support for
      :ref:`include-codeblock command<include-codeblock-command>`.


    Returns:
        str: HTML output translated version of the given file.
    """
    return MdPo2HTML(
        pofiles,
        ignore=ignore,
        po_encoding=po_encoding,
        command_aliases=command_aliases,
        **kwargs,
    ).translate(
        filepath_or_content,
        save=save,
        html_encoding=html_encoding,
    )



"""Custom events executed during the parsing process of an implementation."""

import importlib
import os
import sys


def raise_skip_event(events, event_name, *event_args):
    """Execute all functions defined for an event of a parser.

    If a function returns ``False``, this function will return ``True`` meaning
    that an event is trying to skip the associated function.

    Args:
        events (dict): Dictionary with all events defined.
        event_name (str): Name of the event whose functions will be executed.
        *event_args: Arguments propagated to the events functions.

    Returns:
        bool: ``True`` if an event function returns ``False``, ``False``
        otherwise.
    """
    try:
        pre_events = events[event_name]
    except KeyError:
        return False
    skip = False
    for event in pre_events:
        if event(*event_args) is False:
            skip = True
    return skip


def _block_msg(block, details):
    if details:
        return f'{block.name} - {details}'
    return block.name


def debug_events(program):
    """Debugging events for interfaces. Displays in STDOUT all event targets.

    Args:
        program (str): Command line interface name, to display it at the
            beginning of the output.

    Returns:
        dict: Event target printing functions.
    """
    def debug(event, msg):
        import datetime

        date = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
        sys.stdout.write(f'{program}[DEBUG]::{date}::{event}:: {msg}\n')

    def print_msgid(
            self,  # noqa: ARG001
            msgid,
            msgstr,
            msgctxt,
            tcomment,
            flags,
    ):
        msg = f"msgid='{msgid}'"
        if msgstr:
            msg += f" - msgstr='{msgstr}'"
        if msgctxt:
            msg += f" - msgctxt='{msgctxt}'"
        if tcomment:
            msg += f" - tcomment='{tcomment}'"
        if flags:
            msg += f" - flags='{flags}'"
        debug('msgid', msg)

    def print_command(
            self,  # noqa: ARG001
            mdpo_command,
            comment,
            original_command,
    ):
        msg = mdpo_command
        if comment:
            msg += f' - {comment}'
        if original_command and original_command != mdpo_command:
            msg += f" (original command: '{original_command}')"
        debug('command', msg)

    def print_enter_block(self, block, details):  # noqa: ARG001
        debug('enter_block', _block_msg(block, details))

    def print_leave_block(self, block, details):  # noqa: ARG001
        debug('leave_block', _block_msg(block, details))

    def print_enter_span(self, span, details):  # noqa: ARG001
        debug('enter_span', _block_msg(span, details))

    def print_leave_span(self, span, details):  # noqa: ARG001
        debug('leave_span', _block_msg(span, details))

    def print_text(self, block, text):  # noqa: ARG001
        debug('text', text)

    def print_link_reference(self, target, href, title):  # noqa: ARG001
        msg = f"target='{target}'"
        if href:
            msg += f" - href='{href}'"
        if title:
            msg += f" - title='{title}'"
        debug('link_reference', msg)

    return {
        'msgid': print_msgid,
        'enter_block': print_enter_block,
        'leave_block': print_leave_block,
        'enter_span': print_enter_span,
        'leave_span': print_leave_span,
        'text': print_text,
        'command': print_command,
        'link_reference': print_link_reference,
    }


def add_debug_events(implementation_name, events):
    """Add debugging events to an events dict.

    Args:
        implementation_name (str): Implementation name, shown when
            printing debugging events.
        events (dict): Events dictionary.
    """
    for event_name, function in debug_events(implementation_name).items():
        if event_name not in events:
            events[event_name] = []
        events[event_name].append(function)


def parse_events_kwarg(events_kwarg):
    """Parse ``events`` kwarg passed to implementations.

    Each event can be a function or a path to a function inside a file
    using the syntax ``path/to/file.py::function``.

    Args:
        events_kwarg (dict): Dictionary of event names and their location.

    Returns:
        dict: Event names mapping to a list of functions to execute.
    """
    events = {}
    for event_name, funcs_func_or_filefuncs in events_kwarg.items():
        funcs_or_filefuncs = (
            [funcs_func_or_filefuncs]
            if callable(funcs_func_or_filefuncs)
            or isinstance(funcs_func_or_filefuncs, str)
            else funcs_func_or_filefuncs
        )

        events[event_name] = []
        for func_or_filefunc in funcs_or_filefuncs:
            if isinstance(func_or_filefunc, str):
                # is a string with the syntax 'path/to/file.py::func'
                #
                # import the file, execute as a module and get the function
                if '::' not in func_or_filefunc:
                    raise ValueError(
                        'Function not specified for file'
                        f" '{func_or_filefunc}' defined for event"
                        f" '{event_name}'",
                    )

                fpath, funcname = func_or_filefunc.split('::')
                if not os.path.isfile(fpath):
                    raise FileNotFoundError(
                        f"File '{fpath}' specified for event"
                        f" '{event_name}' not found",
                    )

                modname = os.path.splitext(fpath)[0].replace(os.sep, '.')
                if modname in sys.modules:
                    mod = sys.modules[modname]
                else:
                    spec = importlib.util.spec_from_file_location(
                        modname, fpath,
                    )
                    mod = importlib.util.module_from_spec(spec)
                    sys.modules[modname] = mod
                    spec.loader.exec_module(mod)

                func_owner = mod
                if '.' in funcname:
                    # is a method
                    classname, funcname = funcname.split('.')
                    if not hasattr(mod, classname):
                        raise ValueError(
                            f"Class '{classname}' specified for event"
                            f" '{event_name}' not found in file '{fpath}'",
                        )

                    cls = getattr(mod, classname)
                    if not hasattr(cls, funcname):
                        raise ValueError(
                            f"Method '{funcname}' specified for event"
                            f" '{event_name}' not found in class '{classname}'"
                            f" in file '{fpath}'",
                        )
                    func_owner = cls
                elif not hasattr(mod, funcname):
                    raise ValueError(
                        f"Function '{funcname}' specified for event"
                        f" '{event_name}' not found in file '{fpath}'",
                    )

                events[event_name].append(getattr(func_owner, funcname))
            else:
                # is a function
                events[event_name].append(func_or_filefunc)
    return events



"""Text utilities for mdpo."""

import os
import sys


INFINITE_WRAPWIDTH = 65535


def and_join(values):
    """Comma and space join using "and" between the last and penultimate items.

    Args:
        values (list): Values to join.
    """
    return ', '.join(values[:-1]) + f' and {values[-1]}'


def min_not_max_chars_in_a_row(char, text, default=1):
    r"""Return minimum possible of characters not found in a row for a string.

    For example, given the string ``"c cc cccc"`` and the character ``"c"``,
    returns the minimum number of characters in a row that are not found, so
    ``3`` in this case.

    This function is used in the source code to compute the string that
    wraps markdown code spans. Given the code span
    ``"code that contains 3 \`\`\` and 2 \`\` backticks"`` and the character
    ``"`"``, this function returns  ``1``.

    Args:
        char (str): Character to search.
        text (str): Text inside which find the character repeated in a row.
        default (int): Default value to return if the character is not found.

    Returns:
        int: Minimum number possible of characters not found in a row.
    """
    in_a_rows, _current_in_a_row, _in_the_row = ([], 0, False)
    for ch in text:
        if ch == char:
            _current_in_a_row += 1
            _in_the_row = True
        elif _in_the_row:
            _in_the_row = False
            if _current_in_a_row not in in_a_rows:
                in_a_rows.append(_current_in_a_row)
            _current_in_a_row = 0
    if _in_the_row and _current_in_a_row not in in_a_rows:
        in_a_rows.append(_current_in_a_row)

    if in_a_rows:
        response = None
        for n in range(1, max(in_a_rows) + 2):
            if n not in in_a_rows:
                response = n
                break
    else:
        response = default
    return response


def parse_escaped_pair(value, separator=':'):
    r"""Escapes a pair key-value separated by a character.

    The separator can be escaped using the character '\'.

    Args:
        value (str): String to be converted to a pair key-value.
        separator (str): Separator to use for the split.

    Raises:
        ValueError: The value doesn't contains an unescaped valid separator.

    Returns:
        tuple: Parsed key-value pair.
    """
    import re
    splits = re.split(
        re.compile(fr'([^\\]{separator})'),
        value.lstrip('\\'),
        maxsplit=1,
    )
    if len(splits) == 1:
        raise ValueError()
    return (
        (splits[0] + splits[1][0]).replace(rf'\\{separator}', separator, 1),
        splits[2].lstrip(),
    )


def parse_escaped_pairs(pairs, separator=':'):
    r"""Escapes multiples pairs key-value separated by a character.

    The separator can be escaped using the character '\'.

    Args:
        pairs (list): List of texts to parse.
        separator (str): Separator to use for the splits.

    Raises:
        ValueError: A value doesn't contains an unescaped valid separator.
        KeyError: Repeated keys in pairs.

    Returns:
        dict: Key-value pairs.
    """
    response = {}
    for pair in pairs:
        try:
            key, value = parse_escaped_pair(pair, separator=separator)
        except ValueError as exc:
            raise ValueError(pair) from exc
        if key in response:
            raise KeyError(key)
        response[key] = value
    return response


def parse_strint_0_inf(value):
    """Parse a string to a integer accepting infinte values.

    Converts an integer passed as string in an integer or ``math.inf`` if
    the passed value is ``"0"`` or ``math.inf``.

    Args:
        value (str): Value to parse.
    """
    if isinstance(value, str) and value.lower() == 'nan':
        raise ValueError(f"invalid strict converting of '{value}' to number")

    num = float(value)
    try:
        return int(num) if num > 0 else INFINITE_WRAPWIDTH
    except OverflowError:  # cannot convert float infinity to integer
        return INFINITE_WRAPWIDTH


def parse_wrapwidth_argument(value):
    """Parse the argument ``-w/--wrapwidth``.

    Args:
        value (str): Wrapwidth value.
    """
    try:
        value = parse_strint_0_inf(value)
    except ValueError as exc:
        if os.environ.get('_MDPO_RUNNING'):  # executed as CLI
            sys.stderr.write(
                f"Invalid value '{value}' for -w/--wrapwidth argument.\n",
            )
            sys.exit(1)
        raise ValueError(
            f"Invalid value '{value}' for wrapwidth argument.",
        ) from exc
    return value



#!/usr/bin/env python

"""md2po2md command line interface.

See :ref:`md2po2md CLI<cli:md2po2md>`.
"""

import argparse
import itertools
import sys

from mdpo.cli import (
    SPHINX_IS_RUNNING,
    add_check_option,
    add_command_alias_argument,
    add_common_cli_first_arguments,
    add_debug_option,
    add_encoding_arguments,
    add_extensions_argument,
    add_include_codeblocks_option,
    add_no_empty_msgstr_option,
    add_no_obsolete_option,
    add_nolocation_option,
    add_wrapwidth_argument,
    cli_codespan,
    parse_command_aliases_cli_arguments,
)
from mdpo.io import environ
from mdpo.md2po2md import markdown_to_pofile_to_markdown
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS


DESCRIPTION = (
    'Translates Markdown files using PO files for a set of predefined language'
    ' codes creating multiple directories, one for each language.'
)


def build_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, add_help=False)
    add_common_cli_first_arguments(parser)
    parser.add_argument(
        'input_paths_glob', metavar='GLOB', nargs='*',
        help='Glob to markdown input files to translate.'
             ' If not provided, will be read from STDIN.',
    )
    parser.add_argument(
        '-l', '--lang', dest='langs', default=[], nargs='*',
        action='append',
        help='Language codes used to create the output directories.'
             ' This argument can be passed multiple times. Also, all'
             ' languages can be defined after this argument with'
             f" {cli_codespan('-l es_ES fr_FR de_DE')}.",
        metavar='LANG', required=True,
    )

    output_paths_schema_help = '' if SPHINX_IS_RUNNING else (
        " For example, for the schema 'locale/{lang}', the languages"
        " 'es' and 'fr' and a 'README.md' as input, the next files"
        " will be written: 'locale/es/README.po', 'locale/es/README.md',"
        " 'locale/fr/README.po' and 'locale/fr/README.md'."
        " Note that you can omit '{basename}', specifying a"
        " directory for each language with 'locale/{lang}' for this"
        ' example.'
    )
    parser.add_argument(
        '-o', '--output', dest='output_paths_schema', required=True, type=str,
        help='Path schema for outputs, built using placeholders. There is a'
             ' mandatory placeholder for languages: {lang};and one optional'
             f' for output basename: {{basename}}.{output_paths_schema_help}'
             ' Unexistent directories and files will be created, so you do not'
             ' have to prepare the output directories before the execution.',
        metavar='PATH_SCHEMA',
    )
    add_nolocation_option(parser)
    add_extensions_argument(parser)
    add_command_alias_argument(parser)
    add_wrapwidth_argument(
        parser,
        markup='po',
        markup_prefix=True,
        short=False,
        help_to_render='PO files',
    )
    add_wrapwidth_argument(
        parser, markup='md', markup_prefix=True, short=False, default='80',
    )
    add_include_codeblocks_option(parser)
    add_encoding_arguments(parser)
    add_debug_option(parser)
    add_check_option(parser)
    add_no_obsolete_option(parser)
    add_no_empty_msgstr_option(parser)
    return parser


def parse_options(args):
    parser = build_parser()
    if '-h' in args or '--help' in args:
        parser.print_help()
        sys.exit(1)
    opts, unknown = parser.parse_known_args(args)

    input_paths_glob = ''
    if not sys.stdin.isatty():
        input_paths_glob += sys.stdin.read().strip('\n')
    if isinstance(opts.input_paths_glob, list) and opts.input_paths_glob:
        input_paths_glob += opts.input_paths_glob[0]
    if not input_paths_glob:
        sys.stderr.write('Files or content to translate not specified\n')
        sys.exit(1)
    opts.input_paths_glob = input_paths_glob

    opts.langs = set(itertools.chain(*opts.langs))  # flatten

    if opts.extensions is None:
        opts.extensions = DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS

    opts.command_aliases = parse_command_aliases_cli_arguments(
        opts.command_aliases,
    )

    return opts


def run(args=frozenset()):
    exitcode = 0

    with environ(_MDPO_RUNNING='true'):
        opts = parse_options(args)

        kwargs = {
            'extensions': opts.extensions,
            'command_aliases': opts.command_aliases,
            'debug': opts.debug,
            'location': opts.location,
            'po_wrapwidth': opts.po_wrapwidth,
            'md_wrapwidth': opts.md_wrapwidth,
            'po_encoding': opts.po_encoding,
            'md_encoding': opts.md_encoding,
            'include_codeblocks': opts.include_codeblocks,
            '_check_saved_files_changed': opts.check_saved_files_changed,
        }

        (
            _saved_files_changed,
            obsoletes,
            empty,
        ) = markdown_to_pofile_to_markdown(
            opts.langs,
            opts.input_paths_glob,
            opts.output_paths_schema,
            **kwargs,
        )
        if opts.check_saved_files_changed and _saved_files_changed:
            exitcode = 2

        if opts.no_obsolete and obsoletes:
            exitcode = 3

            if not opts.quiet:
                sys.stderr.write(
                    "Obsolete messages found at PO files and"
                    " passed '--no-obsolete'\n",
                )

        if opts.no_empty_msgstr and empty:
            exitcode = 4

            if not opts.quiet:
                sys.stderr.write(
                    "Empty msgstr found at PO files and"
                    " passed '--no-empty-msgstr'\n",
                )

    return exitcode


def main():
    raise SystemExit(run(args=sys.argv[1:]))  # pragma: no cover


if __name__ == '__main__':
    main()



"""md4c related stuff for mdpo."""


#: :list: `md4c parser <https://github.com/mity/md4c>`_ extensions
#: used by default on :doc:`md2po </dev/reference/mdpo.md2po>` and
#: :doc:`po2md </dev/reference/mdpo.po2md>` implementations.


DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS = [
    'collapse_whitespace',
    'tables',
    'strikethrough',
    'tasklists',
    'latex_math_spans',
    'wikilinks',
]

#: :dict: mapping from `pymd4c <https://pymd4c.dcpx.org/>`_
#: block type values to block readable names
READABLE_BLOCK_NAMES = {
    1: 'quote',
    2: 'unordered list',
    3: 'ordered list',
    6: 'header',
    7: 'code',
    8: 'html',
    9: 'paragraph',
    10: 'table',
}



#!/usr/bin/env python

"""md2po command line interface.

See :ref:`md2po CLI<cli:md2po>`.
"""

import argparse
import sys

from mdpo.cli import (
    CLOSE_QUOTE_CHAR,
    OPEN_QUOTE_CHAR,
    add_check_option,
    add_command_alias_argument,
    add_common_cli_first_arguments,
    add_debug_option,
    add_encoding_arguments,
    add_event_argument,
    add_extensions_argument,
    add_include_codeblocks_option,
    add_no_empty_msgstr_option,
    add_no_obsolete_option,
    add_nolocation_option,
    add_wrapwidth_argument,
    cli_codespan,
    parse_command_aliases_cli_arguments,
    parse_event_argument,
    parse_metadata_cli_arguments,
)
from mdpo.io import environ
from mdpo.md2po import Md2Po
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS


DESCRIPTION = (
    'Utility like xgettext to extract Markdown contents dumping them'
    ' inside PO files.'
)


def build_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, add_help=False)
    add_common_cli_first_arguments(parser)
    parser.add_argument(
        'files_or_content', metavar='GLOBS_FILES_OR_CONTENT',
        nargs='*',
        help='Globs to markdown input files, paths to files or Markdown'
             ' content. If not provided, will be read from STDIN.',
    )
    parser.add_argument(
        '-i', '--ignore', dest='ignore', default=[], action='append',
        help='Path to a file to ignore. This argument can be passed multiple'
             ' times.',
        metavar='PATH',
    )
    parser.add_argument(
        '-p', '--po-filepath', '--pofilepath', dest='po_filepath',
        default=None,
        help='Merge new msgids in the po file indicated at this parameter (if'
             f' {cli_codespan("--save")} argument is passed) or use the msgids'
             ' of the file as reference for mark not found as obsoletes if'
             f' {cli_codespan("--merge-pofiles")} parameter is not passed.',
        metavar='OUTPUT_PO_FILEPATH',
    )
    parser.add_argument(
        '-s', '--save', dest='save', action='store_true',
        help='Save new found msgids to the po file indicated as parameter'
             f' {cli_codespan("--po-filepath")}. Passing this option without'
             f' defining the argument {cli_codespan("--po-filepath")} will'
              ' raise an error.',
    )
    parser.add_argument(
        '--mo-filepath', '--mofilepath', dest='mo_filepath',
        default=None,
        help='The resulting PO file will be compiled to a mofile and saved in'
             ' the path specified at this parameter.',
        metavar='OUTPUT_MO_FILEPATH',
    )
    parser.add_argument(
        '--plaintext', dest='plaintext', action='store_true',
        help='Do not include markdown markup characters in extracted msgids'
             f' for {cli_codespan("**bold text**", cli=False)},'
             f' {cli_codespan("*italic text*", cli=False)},'
             f' {cli_codespan("``inline code``")} and'
             f' {cli_codespan("[link](target)")}.',
    )
    add_wrapwidth_argument(parser, markup='po')
    parser.add_argument(
        '-m', '--merge-po-files', '--merge-pofiles',
        dest='mark_not_found_as_obsolete',
        action='store_false',
        help='Messages not found which are already stored in the PO file'
             f' passed as {cli_codespan("--po-filepath")} argument will not be'
             ' marked as obsolete.',
    )
    parser.add_argument(
        '-r', '--remove-not-found',
        dest='preserve_not_found',
        action='store_false',
        help='Messages not found which are already stored in the PO file'
             f' passed as {cli_codespan("--po-filepath")} parameter will be'
             ' removed. Only has effect used in combination with'
             f' {cli_codespan("--merge-pofiles")}. If you pass this option,'
             f' {cli_codespan("--merge-po-files")} will be ignored.',
    )
    add_nolocation_option(parser)
    add_extensions_argument(parser)
    add_encoding_arguments(
        parser,
        po_encoding_help='Resulting PO file encoding.',
    )
    parser.add_argument(
        '-a', '--xheader', dest='xheader', action='store_true',
        help='Include in the resulting PO file the mdpo specification'
             ' X-Header "X-Generation", whose value is "mdpo v<version>".',
    )
    add_include_codeblocks_option(parser)
    parser.add_argument(
        '--ignore-msgids', dest='ignore_msgids', default=None,
        help='Path to a plain text file where all msgids to ignore from being'
             ' extracted are located, separated by newlines.',
    )

    # patch for sphinx-argparse-cli compatibility (use Unicode quotation marks)
    example_codespan = (
        f'-d {OPEN_QUOTE_CHAR}Content-Type: text/plain;'
        f' charset=utf-8{CLOSE_QUOTE_CHAR}'
        f' -d {OPEN_QUOTE_CHAR}Language: es{CLOSE_QUOTE_CHAR}'
    )
    metadata_help_example = (
        ' For example, to define UTF-8 encoding and Spanish language use'
        f' {cli_codespan(example_codespan)}.'
    )
    parser.add_argument(
        '-d', '--metadata', dest='metadata', default=[], action='append',
        metavar='Key:Value',
        help='Custom metadata key-value pairs to include in the produced'
             ' PO file. This argument can be passed multiple times.'
             ' If the file contains previous metadata fields, these will'
             ' be updated preserving the values of the already defined.'
             f'{metadata_help_example}',
    )

    add_command_alias_argument(parser)
    add_event_argument(parser)
    add_debug_option(parser)
    add_check_option(parser)
    add_no_obsolete_option(parser)
    add_no_empty_msgstr_option(parser)
    return parser


def parse_options(args):
    parser = build_parser()
    if '-h' in args or '--help' in args:
        parser.print_help()
        sys.exit(1)
    opts, _ = parser.parse_known_args(args)

    files_or_content = ''
    if not sys.stdin.isatty():
        files_or_content += sys.stdin.read().strip('\n')
    if isinstance(opts.files_or_content, list) and opts.files_or_content:
        if len(opts.files_or_content) == 1:
            files_or_content += opts.files_or_content[0]
        else:
            files_or_content = opts.files_or_content
    if not files_or_content:
        sys.stderr.write('Files or content to extract not specified\n')
        sys.exit(1)
    opts.files_or_content = files_or_content

    if opts.extensions is None:
        opts.extensions = DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS

    if opts.ignore_msgids is not None:
        with open(opts.ignore_msgids, encoding=opts.po_encoding) as f:
            opts.ignore_msgids = f.read().splitlines()
    else:
        opts.ignore_msgids = []

    opts.command_aliases = parse_command_aliases_cli_arguments(
        opts.command_aliases,
    )
    opts.events = parse_event_argument(opts.events)
    opts.metadata = parse_metadata_cli_arguments(
        opts.metadata,
    )

    return opts


def run(args=frozenset()):
    with environ(_MDPO_RUNNING='true'):
        opts = parse_options(args)

        init_kwargs = {
            'ignore': opts.ignore,
            'plaintext': opts.plaintext,
            'mark_not_found_as_obsolete': opts.mark_not_found_as_obsolete,
            'preserve_not_found': opts.preserve_not_found,
            'location': opts.location,
            'extensions': opts.extensions,
            'xheader': opts.xheader,
            'include_codeblocks': opts.include_codeblocks,
            'ignore_msgids': opts.ignore_msgids,
            'command_aliases': opts.command_aliases,
            'metadata': opts.metadata,
            'events': opts.events,
            'debug': opts.debug,
            '_check_saved_files_changed': opts.check_saved_files_changed,
        }

        extract_kwargs = {
            'po_filepath': opts.po_filepath,
            'save': opts.save,
            'mo_filepath': opts.mo_filepath,
            'po_encoding': opts.po_encoding,
            'md_encoding': opts.md_encoding,
            'wrapwidth': opts.wrapwidth,
        }

        md2po = Md2Po(opts.files_or_content, **init_kwargs)
        pofile = md2po.extract(**extract_kwargs)
        exitcode = 0

        if not opts.quiet:
            sys.stdout.write(f'{pofile.__unicode__()}\n')

        # pre-commit mode
        if opts.check_saved_files_changed and md2po._saved_files_changed:
            exitcode = 2

        if opts.no_obsolete and md2po.obsoletes:
            if not opts.quiet:
                sys.stderr.write(
                    (
                        f"Obsolete messages found at {opts.po_filepath}"
                        " and passed '--no-obsolete'\n"
                    ),
                )
            exitcode = 3

        if opts.no_empty_msgstr:
            for entry in pofile:
                if not entry.msgstr:
                    if not opts.quiet:
                        sys.stderr.write(
                            (
                                f"Empty msgstr found at {opts.po_filepath}"
                                " and passed '--no-empty-msgstr'\n"
                            ),
                        )
                        exitcode = 4
                    break

    return (pofile, exitcode)


def main():
    raise SystemExit(run(args=sys.argv[1:])[1])  # pragma: no cover


if __name__ == '__main__':
    main()
